;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; This is a 430eForth assembler listing based on the original script from
; Dr. Chen Hanson Ting as described in his book: Zen and the Forth Language:
; EFORTH for the MSP430 from Texas Instruments (Kindle Edition). It was first
; written by him in IAR Assembler, then transferred to the CCS and now adapted
; by Michael Kalus for the naken_asm.
;
; Some Flash Tools and support for WORDLISTs, VOCABULARY Prefixes, ITEMs and
; STICKY Words was added by Manfred Mahlow.
;
; MIT License
; ----------------------------------------------------------------------------
; Copyright (c) 2014 Dr. Chen-Hanson Ting  CCS Version    430eForth4.3
;           (c) 2018 Michael Kalus         Naken Version  430eForth4.3n
;           (c) 2018 Manfred Mahlow          Flash Tools  430eForth4.3n1
;           (c) 2019 Manfred Mahlow          430eForth-lp2553-43n5vis430eForth
;
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
;
; The above copyright notice and this permission notice shall be included in all
; copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; 7/7/2012 430eForth1.0, from eForth86.asm and 430uForth
; 7/4/2012 Move 430uForth2.1 from IAR to CCS 5.2
; 8/5/2014 Move 430eForth2.2 to CCS 6.0.  Fix linkage of OVER.
;	Software UART at 2400 baud.
; 8/10/2014 430eForth2.3 9600 baud, thanks to Dirk Bruehl and
;	Michael Kalus of www.4e4th.org
; 8/10/2014 430eForth2.4 Restore ERASE and WRITE
; 8/20/2014 430eForth2.5 Test Segment D
; 8/25/2014 430eForth2.6 Turnkey
; 8/26/2014 430eForth2.7 Optimize
; 9/16/2014 430eForth3.1 Tail recursion, APP!
; 10/11/2014 430eForth4.1 Direct thread, more optimization
; 10/23/2014 430eForth4.2 Direct thread, pack lists
; 11/12/2014 430eForth4.2 Direct thread, final
;
; Build for and verified on MSP430G2 LaunchPad from TI
; Assembled with Code Composer Sudio 6.0 IDE
; Internal DCO at 8 MHz
; Hardware UART at 9600 baud. TXD and RXD must be crossed.

;CCS:   ;ting
;	.nolist 
;	.title "msp430 eForth 4.3" 
;	.cdecls C,LIST,"msp430g2553.h"  ; Include device header file 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; 05/13/2018 Moved 430eForth4.3 from CSS to Michael Kohn's naken_asm (ver.
;            23 april 2018) - ok                                           ;mk
; 20180624 $," - bug fixed. Thanks to Manfred Mahlow.                      ;mk
; 20180628 DIGIT? : bug fix, 0= 0> were handled as numbers                 ;MM
; 20180629 FSCAN added, sets CP to the lowest free flash addr at BOOT time.
;          FSCAN is executed before COLD executes QUIT.
;          Version string changed in HI from 43n to 43n1.                  ;MM
; 20180630 ERASE and WRITE renamed to IERASE IWRITE due to name conflict   ;MM
; 20180701 LITERAL and ALIGNED revealed.                                   ;MM
; 20180707 Flash Test QFLASH ( a -- a ) added. Aborts with message ?flash
;          if a > EDM. (EDM = End of Dictionary Memory space in the flash) ;MM
; 20180720 Support for tags based wordlists and VOCs added. Default search
;          order is FORTH ROOT. VOCs search order is <VOC> ROOT. ( new: ROOT
;          FORTH DEFINITIONS VOC )                                         ;MM
; 20180730 Implicit context switching added ( new: CASTED )                ;MM
; 20180826 CASTED renamed to ITEM and moved to a Forth source code file    ;MM
; 20181120 FORTH and ROOT are vocabularies now. vROOT added for VOCS       ;MM
; 20181124 Context Switching is done after a word is executed or compiled. ;MM
; 20181207 FORTH and ROOT are VOCs, vROOT removed,FIRST added              ;MM
; 20190106 VOCs are no longer implemented as ITEMS.
;          ITEM and STICKY added again.                                    ;MM
; 20190131 Macros added to create headers and tags                         ;MM
; 20190208 DOLIT,0 replaced with constant ZERO                             ;MM
; 20190208 DOLIT,1,ANDD replaced with GETB0                                ;MM
; 20190314 Bug fix for UM/MOD anf TOKEN                                    ;MM
; 20190330 FIRST removed ALSO added                                        ;MM
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;naken: ;mk
    .msp430 
    .include "msp430g2553.inc"  ; MCU-specific register equates for naken_asm

    .include "header.inc"       ; adding header macros  MM-190130

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Direct Thread Model of eForth

; CCS: .equ ; naken: equ   ;mk
;;	CPU registers
tos	equ	R4
stack   equ	R5
ip	equ	R6
temp0	equ	R7
temp1	equ	R8
temp2	equ	R9
temp3	equ	R10

;; Macros
; CCS: <name> .macro ; naken: .macro <name>   ;mk
.macro pops	;DROP
	mov.w	@stack+,tos
    .endm

.macro pushs ;DUP
	decd.w	stack
	mov.w	tos,0(stack)
    .endm;; Constants

.macro INEXT ;mk renamed (dollar)NEXT to INEXT - inline code for NEXT.
	mov @ip+,pc	; fetch code address into PC
    .endm

.macro INEST ;mk renamed (dollar)NEST to INEST - inline code for NEST.
	.align	16  ; CCS: 2 bytes align ; naken: 16 bit align. mk
	call	#DOLST	; fetch code address into PC, W=PFA
    .endm

.macro ICONST ;mk renamed (dollar)CONST to ICONST - inline code calling DOCON.
	.align	16  ; CCS: 2 bytes align ; naken: 16 bit align. mk
	call	#DOCON	; fetch code address into PC, W=PFA
	.endm

;; Assembler constants

XTAG   equ  020H  ;lexicon tag bit                            MM++180712
COMPO	 equ	040H	;lexicon compile only bit
IMEDD	 equ	080H	;lexicon immediate bit
MASKK	 equ	07F1FH	;lexicon bit mask
CELLL	 equ	2	;size of a cell
BASEE	 equ	10	;default radix
VOCSS	 equ	8	;depth of vocabulary stack
BKSPP	 equ	8	;backspace
LF	   equ	10	;line feed
CRR	   equ	13	;carriage return
ERR	   equ	27	;error escape
TIC	   equ	39	;tick
CALLL	 equ	012B0H	;NOP CALL opcodes

UPP	   equ	200H
DPP	   equ	220H
SPP	   equ	378H	;data stack
TIBB   equ	380H	;terminal input buffer
RPP	   equ	3F8H	;return stack
CODEE	 equ	0C000H	;code dictionary
COLDD	 equ	0FFFEH	;cold start vector
EM	   equ	0FFFFH	;top of memory

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.list
;;;;
;; Main entry points and COLD start data

;	.text       ; CCS: Predefined Memory Segment Name : Main memory (flash or ROM)
    .org 0C000H ; naken : MSP430G2553 main memory : C000-FFFF = 16KB flash ROM 
                ;                                   FFE0-FFFF = interrupt vectors

;	.global	main ; CCS: File Reference Directive 
                 ; unused in naken

main: 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.set link = 0H    ; MM++190130

;; Device dependent I/O

; ?KEY	( -- F | c T )
;	Return input character.
  HEAD( 4,"?KEY" ) 
QKEY:
	pushs
QKEY1:
	BIT.B   #UCA0RXIFG,&IFG2
	JZ	FALSE	;return false flag
	MOV.B   &UCA0RXBUF,tos	; read character into TOS
	pushs
	jmp	TRUE

; KEY	( -- c )
;	Return input character.
  HEAD( 3,"KEY" )
KEY:
	pushs
KEY1:
	BIT.B   #UCA0RXIFG,&IFG2
	JZ	KEY1
	MOV.B   &UCA0RXBUF,tos	; read character into TOS
	INEXT

; EMIT	( c -- )
;	Send character c to the output device.
  HEAD( 4,"EMIT" )
EMIT:
EMIT1:
	BIT.B   #UCA0TXIFG,&IFG2
	JZ	EMIT1
	MOV.B   tos,&UCA0TXBUF
	pops
	INEXT

; !IO	( -- )
;	Initialize the serial I/O devices.
;	.dw	EMIT-6
;	.db 3,"!IO"
; HEAD( 3,"!IO" )
STOIO:
; 8MHz
	mov.b   &CALBC1_8MHZ, &BCSCTL1   ; Set DCO
	mov.b   &CALDCO_8MHZ, &DCOCTL	; to 8 MHz.
	mov.b   #006h, &P1SEL	; Use P1.1/P1.2 for USCI_A0
	mov.b   #006h, &P1SEL2	; Use P1.1/P1.2 for USCI_A0
; Configure UART (Koch)
	bis.b   #UCSSEL_2,&UCA0CTL1	;db2 SMCLK
	mov.b   #65,&UCA0BR0	;db3 8MHz 9600 Insgesamt &833 = $341
	mov.b   #3,&UCA0BR1	;db4 8MHz 9600
	mov.b   #UCBRS_2,&UCA0MCTL	;db5 Modulation UCBRSx = 2
	bic.b   #UCSWRST,&UCA0CTL1	;db6 **Initialize USCI
	INEXT	;called from COLD

;; The kernel

; doLIT	( -- w )
;	Push an inline literal.
; HEAD( COMPO+5,"doLIT" )
DOLIT:
	pushs
	mov	@ip+,tos
	INEXT

; doCON	( -- a )
;	Run time routine for CONSTANT, VARIABLE and CREATE.
; HEAD( COMPO+5,"doCON" )
DOCON:
	pushs
	pop	tos
	mov	@tos,tos
	INEXT

; doLIST	( -- )
;	Process colon list..
;	HEAD( 6,"doLIST" )
DOLST:
	mov	ip,temp0	;exchange pointers
	pop	ip	;push return stack
	push	temp0	;restore the pointers
	INEXT

; EXIT	( -- )
;	Terminate a colon definition.
  HEAD( 4,"EXIT" )
EXIT:
	mov	@sp+,ip
	INEXT

; EXECUTE	( ca -- )
;	Execute the word at ca.
	HEAD( 7,"EXECUTE" )
EXECU:
	mov	tos,temp0
	pops
	br	temp0

; @EXECUTE	( a -- )
;	Execute vector stored in address a.
	HEAD( 8,"@EXECUTE" )
ATEXE:
	mov	@tos,temp0
	pops
	br	temp0

; branch	( -- )
;	Branch to an inline address.
;	HEAD( COMPO+6,"branch" )
BRAN:
	mov	@ip+,ip
	INEXT

; ?branch	( f -- )
;	Branch if flag is zero.
;	HEAD( COMPO+7,"?branch" )
QBRAN:
	tst	tos
	pops
	jz	BRAN
	jmp	SKIP

; next	( -- )
;	Run time code for the single index loop.
;	: next ( -- ) \ hilevel model
;	r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;
;	HEAD( COMPO+4,"next" )
DONXT:
	dec	0(sp)   ;decrement index
	jge	BRAN	;loop back
	incd.w	sp	;discard index
SKIP:
	incd.w	ip	;exit loop
	INEXT

; !	( w a -- )
;	Pop the data stack to memory.
	HEAD( 1,"!" )
STORE:
	mov.w	@stack+,0(tos)
	pops
	INEXT

; @	( a -- w )
;	Push memory location to the data stack.
	HEAD( 1,"@" )
AT:
	mov.w	@tos,tos
	INEXT

; C!	( c b -- )
;	Pop the data stack to byte memory.
	HEAD( 2,"C!" )
CSTOR:
	mov.b	@stack+,0(tos)
	inc	stack
	pops
	INEXT

; C@	( b -- c )
;	Push byte memory location to the data stack.
	HEAD( 2,"C@" )
CAT:
	mov.b	@tos,tos
	INEXT

; RP! ( -- )
;	init return stack pointer.
;	HEAD( 3,"RP!" )
RPSTO:
	mov	#RPP,SP	;init return stack
	INEXT

; R> ( -- w )
;	Pop the return stack to the data stack.
  HEAD( 2,"R>" )
RFROM:
	pushs
	pop	tos
	INEXT

; R@ ( -- w )
;	Copy top of return stack to the data stack.
	HEAD( 2,"R@" )
RAT:
	pushs
	mov	0(sp),tos
	INEXT

; >R ( w -- )
;	Push the data stack to the return stack.
	HEAD( COMPO+2,">R" )
TOR:
	push	tos
	pops
	INEXT

; SP! ( -- )
;	Init data stack pointer.
;	HEAD( 3,"SP!" )
SPSTO:
	mov	#SPP,stack	;init parameter stack
	clr	tos
	INEXT

; DROP ( w -- )
;	Discard top stack item.
	HEAD( 4,"DROP" )
DROP:
	pops
	INEXT

; DUP ( w -- w w )
;	Duplicate the top stack item.
	HEAD( 3,"DUP" )
DUPP:
	pushs
	INEXT

; SWAP ( w1 w2 -- w2 w1 )
;	Exchange top two stack items.
	HEAD( 4,"SWAP" )
SWAP:
	mov.w	tos,temp0
	mov.w	@stack,tos
	mov.w	temp0,0(stack)
	INEXT

; OVER ( w1 w2 -- w1 w2 w1 )
;	Copy second stack item to top.
	HEAD( 4,"OVER" )
OVER:
	mov.w	@stack,temp0
	pushs
	mov.w	temp0,tos
	INEXT

; 0< ( n -- t )
;	Return true if n is negative.
	HEAD( 2,"0<" )
ZLESS:
	tst	tos
	jn	TRUE
FALSE:
	clr	tos
	INEXT
TRUE:
	mov	#0x-1,tos
	INEXT

; AND ( w w -- w )
;	Bitwise AND.
	HEAD( 3,"AND" )
ANDD:
	and	@stack+,tos
	INEXT

; OR ( w w -- w )
;	Bitwise inclusive OR.
	HEAD( 2,"OR" )
ORR:
	bis	@stack+,tos
	INEXT

; XOR ( w w -- w )
;	Bitwise exclusive OR.
	HEAD( 3,"XOR" )
XORR:
	xor	@stack+,tos
	INEXT

; UM+ ( w w -- w cy )
;	Add two numbers, return the sum and carry flag.
	HEAD( 3,"UM+" )
UPLUS:
	add	@stack,tos
	mov	tos,0(stack)
	clr	tos
	rlc	tos
	INEXT

;; Common functions

; ?DUP ( w -- w w | 0 )
;	Dup tos if its is not zero.
	HEAD( 4,"?DUP" )
QDUP:
	tst	tos
	jnz	DUPP
	INEXT

; ROT ( w1 w2 w3 -- w2 w3 w1 )
;	Rot 3rd item to top.
	HEAD( 3,"ROT" )
ROT:
	mov.w	0(stack),temp0
	mov.w	tos,0(stack)
	mov.w	2(stack),tos
	mov.w	temp0,2(stack)
	INEXT

; 2DROP ( w w -- )
;	Discard two items on stack.
	HEAD( 5,"2DROP" )
DDROP:
	incd.w	stack
	pops
	INEXT

; 2DUP ( w1 w2 -- w1 w2 w1 w2 )
;	Duplicate top two items.
	HEAD( 4,"2DUP" )
DDUP:
	mov.w	@stack,temp0
	pushs
	decd.w	stack
	mov.w	temp0,0(stack)
	INEXT

; + ( w w -- sum )
;	Add top two items.
	HEAD( 1,"+" )
PLUS:
	add	@stack+,tos
	INEXT

; D+ ( d d -- d )
;	Double addition, as an example using UM+.
	HEAD( 2,"D+" )
DPLUS:
	mov.w	@stack+,temp0
	mov.w	@stack+,temp1
	add.w	temp0,0(stack)
	addc	temp1,tos
	INEXT

; NOT	( w -- w )
;	One's complement of tos.
	HEAD( 3,"NOT" )
INVER:
	inv	tos
	INEXT

; NEGATE ( n -- -n )
;	Two's complement of tos.
	HEAD( 6,"NEGATE" )
NEGAT:
	inv	tos
	inc	tos
	INEXT

; DNEGATE ( d -- -d )
;	Two's complement of top double.
	HEAD( 7,"DNEGATE" )
DNEGA:
	inv	tos
	inv 0(stack)
	inc	0(stack)
	addc	#0,tos
	INEXT

; - ( n1 n2 -- n1-n2 )
;	Subtraction.
	HEAD( 1,"-" )
SUBB:
	sub	@stack+,tos
	jmp	NEGAT

; ABS ( n -- n )
;	Return the absolute value of n.
	HEAD( 3,"ABS" )
ABSS:
	tst.w	tos
	jn	NEGAT
	INEXT

; = ( w w -- t )
;	Return true if top two are equal.
	HEAD( 1,"=" )
EQUAL:
	xor	@stack+,tos
	jnz	FALSE
	jmp	TRUE

; U< ( u u -- t )
;	Unsigned compare of top two items.
	HEAD( 2,"U<" )
ULESS:
	mov	@stack+,temp0
	cmp	tos,temp0
	subc	tos,tos
	INEXT

; < ( n1 n2 -- t )
;	Signed compare of top two items.
	HEAD( 1,"<" )
LESS:
	cmp	@stack+,tos
	jz	FALSE
	jge	TRUE
	jmp	FALSE

; > ( n1 n2 -- t )
;	Signed compare of top two items.
	HEAD( 1,">" )
GREAT:
	cmp	@stack+,tos
	jge	FALSE
	jmp	TRUE

; MAX ( n n -- n )
;	Return the greater of two top stack items.
	HEAD( 3,"MAX" )
MAX:
	cmp	0(stack),tos
MAX1:
	jl	DROP
	incd.w	stack
	INEXT

; MIN ( n n -- n )
;	Return the smaller of top two stack items.
	HEAD( 3,"MIN" )
MIN:
	cmp	tos,0(stack)
	jmp	MAX1


;; Divide

.if 0  ; this code is buggy, fails for ud > 16383                ;MM-190314

; UM/MOD ( udl udh u -- ur uq )
;	Unsigned divide of a double by a single. Return mod and quotient.
	HEAD( 6,"UM/MOD" )
UMMOD:
	mov	tos,temp0
	pops
	mov	#17,temp1
UMMOD2:
	cmp	temp0,tos
	jnc	UMMOD3
	sub	temp0,tos
	setc
	jmp	UMMOD4
UMMOD3:
	clrc
UMMOD4:
	rlc	0(stack)
	rlc	tos
	dec	temp1
	jnz	UMMOD2
	rra	tos
	mov	tos,temp0
	mov	0(stack),tos
	mov	temp0,0(stack)
	INEXT

.else  ; bug fix MM-190314

;   UM/MOD	( udl udh un -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.
	HEAD( 6,"UM/MOD" )
UMMOD:
  mov.w	@stack+,temp0
  mov @stack,temp1   
  call #UMMOSR
  mov temp0,0(stack)
  mov temp1,tos
  INEXT

UMMOSR:
  cmp tos,temp0
  jnc UMMO1
  mov #-1,temp0
  mov temp0,temp1
  ret
UMMO1:
  mov #1,temp2
UMMO2:
  add temp1,temp1
  addc temp0,temp0
  jc UMMO3
  cmp tos,temp0
  jnc UMMO4
UMMO3:
  sub tos,temp0
  add #1,temp1
UMMO4:
  add temp2,temp2
  jnc UMMO2
  ret

.endif


; M/MOD ( d n -- r q )
;	Signed floored divide of double by single. Return mod and quotient.
	HEAD( 5,"M/MOD" )
MSMOD:
	INEST
	.dw	DUPP,ZLESS,DUPP,TOR,QBRAN,MMOD1
	.dw	NEGAT,TOR,DNEGA,RFROM
MMOD1:
	.dw	TOR,DUPP,ZLESS,QBRAN,MMOD2
	.dw	RAT,PLUS
MMOD2:
	.dw	RFROM,UMMOD,RFROM,QBRAN,MMOD3
	.dw	SWAP,NEGAT,SWAP
MMOD3:
	.dw	EXIT

; /MOD ( n n -- r q )
;	Signed divide. Return mod and quotient.
	HEAD( 4,"/MOD" )
SLMOD:
	INEST
	.dw	OVER,ZLESS,SWAP,MSMOD,EXIT

; MOD ( n n -- r )
;	Signed divide. Return mod only.
	HEAD( 3,"MOD" )
MODD:
	INEST
	.dw	SLMOD,DROP,EXIT

; / ( n n -- q )
;	Signed divide. Return quotient only.
	HEAD( 1,"/" )
SLASH:
	INEST
	.dw	SLMOD,SWAP,DROP,EXIT

;; Multiply

; UM* ( u u -- ud )
;	Unsigned multiply. Return double product.
	HEAD( 3,"UM*" )
UMSTA:
	clr	temp0
	mov	#16,temp1
UMSTA2:
	bit	#1,0(stack)
	jz	UMSTA3
	add	tos,temp0
	jmp	UMSTA4
UMSTA3:
	clrc
UMSTA4:
	rrc	temp0
	rrc	0(stack)
	dec	temp1
	jnz	UMSTA2
	mov	temp0,tos
	INEXT

; * ( n n -- n )
;	Signed multiply. Return single product.
	HEAD( 1,"*" )
STAR:
	INEST
	.dw	UMSTA,DROP,EXIT

; M* ( n n -- d )
;	Signed multiply. Return double product.
	HEAD( 2,"M*" )
MSTAR:
	INEST
	.dw	DDUP,XORR,ZLESS,TOR
	.dw	ABSS,SWAP,ABSS,UMSTA,RFROM
	.dw	QBRAN,MSTA1
	.dw	DNEGA
MSTA1:
	.dw	EXIT

; */MOD ( n1 n2 n3 -- r q )
;	Multiply n1 and n2, then divide by n3. Return mod and quotient.
	HEAD( 5,"*/MOD" )
SSMOD:
	INEST
	.dw	TOR,MSTAR,RFROM,MSMOD,EXIT

; */ ( n1 n2 n3 -- q )
;	Multiply n1 by n2, then divide by n3. Return quotient only.
	HEAD( 2,"*/" )
STASL:
	INEST
	.dw	SSMOD,SWAP,DROP,EXIT

;; Miscellaneous

; 1+ ( a -- a+1 )
;	Increment.
;	HEAD( 2,"1+" )
ONEP:
	add	#1,tos
	INEXT

; 1- ( a -- a-1 )
;	Decrement
;	HEAD( 2,"1-" )
ONEM:
	sub	#1,tos
	INEXT

; 2+ ( a -- a+2 )
;	Add cell size in byte to address.
;	HEAD( 2,"2+" )
CELLP:
	add	#2,tos
	INEXT

; 2- ( a -- a-2 )
;	Subtract cell size in byte from address.
;	HEAD( 2,"2-" )
CELLM:
	sub	#2,tos
	INEXT

; 2* ( n -- 2*n )
;	Multiply tos by cell size in bytes.
	HEAD( 2,"2*" )
CELLS:
	rla	tos
	INEXT

; 2/ ( n -- n/2 )
;	Divide tos by cell size in bytes.
	HEAD( 2,"2/" )
TWOSL:
	rra	tos
	INEXT

; ALIGNED ( b -- a )
;	Align address to the cell boundary.
	HEAD( 7,"ALIGNED" )
ALGND:
	add	#1,tos
	bic	#1,tos
	INEXT

; >CHAR ( c -- c )
;	Filter non-printing characters.
;	HEAD( 5,">CHAR" )
TCHAR:
	INEST
	.dw	BLANK,MAX	;mask msb
	.dw	DOLIT,126,MIN	;check for printable
	.dw	EXIT

; DEPTH ( -- n )
;	Return the depth of the data stack.
	HEAD( 5,"DEPTH" )
DEPTH:
	mov	stack,temp0
	pushs
	mov	#SPP,tos
	sub	temp0,tos
	rra	tos
	INEXT

; PICK ( ... +n -- ... w )
;	Copy the nth stack item to tos.
	HEAD( 4,"PICK" )
PICK:
	rla	tos
	add stack,tos
	mov	@tos,tos
	INEXT

;; Memory access

; +! ( n a -- )
;	Add n to the contents at address a.
	HEAD( 2,"+!" )
PSTOR:
	add @stack+,0(tos)
	pops
	INEXT

; COUNT ( b -- b +n )
;	Return count byte of a string and add 1 to byte address.
	HEAD( 5,"COUNT" )
COUNT:
	mov.b	@tos+,temp0
	pushs
	mov	temp0,tos
	INEXT

; CMOVE ( b1 b2 u -- )
;	Copy u bytes from b1 to b2.
	HEAD( 5,"CMOVE" )
CMOVE:
	mov	@stack+,temp0	;destination
	mov	@stack+,temp1	;source
	jmp	CMOVE2
CMOVE1:
	mov.b	@temp1+,0(temp0)
	inc	temp0
CMOVE2:
	dec	tos
	jn	CMOVE3	;I need a jp.  Oh, well.
	jmp	CMOVE1
CMOVE3:
	JMP	DROP

; FILL ( b u c -- )
;	Fill u bytes of character c to area beginning at b.
	HEAD( 4,"FILL" )
FILL:
	mov	@stack+,temp0	;count
	mov	@stack+,temp1	;destination
	jmp	FIL2
FIL1:
	mov.b	tos,0(temp1)
	inc	temp1
FIL2:
	dec	temp0
	jn	FIL3
	jmp	FIL1
FIL3:
	JMP	DROP

;; User variables

; 'BOOT ( -- a )
;	The application startup vector.
	HEAD( 5,"'BOOT" )
TBOOT:
	ICONST
	.dw	200H

; BASE ( -- a )
;	Storage of the radix base for numeric I/O.
	HEAD( 4,"BASE" )
BASE:
	ICONST
	.dw	202H

; tmp ( -- a )
;	A temporary storage location used in parse and find.
;	HEAD( COMPO+3,"tmp" )
TEMP:
	ICONST
	.dw	204H

; #TIB ( -- a )
;	Hold the character pointer while parsing input stream.
;	HEAD( 4,"#TIB" )
NTIB:
	ICONST
	.dw	206H

; >IN ( -- a )
;	Hold the character pointer while parsing input stream.
;	HEAD( 3,">IN" )
INN:
	ICONST
	.dw	208H

; HLD ( -- a )
;	Hold a pointer in building a numeric output string.
;	HEAD( 3,"HLD" )
HLD:
	ICONST
	.dw	20AH

; 'EVAL ( -- a )
;	A area to specify vocabulary search order.
;	HEAD( 5,"'EVAL" )
TEVAL:
	ICONST
	.dw	20CH

; CONTEXT ( -- a )
;	A area to specify vocabulary search order.
;	HEAD( 7,"CONTEXT" )
CNTXT:
	ICONST
	.dw	20EH

; CP ( -- a )
;	Point to the top of the code dictionary.
	HEAD( 2,"CP" )
CP:
	ICONST
	.dw	210H

; DP ( -- a )
;	Point to the bottom of the free ram area.
	HEAD( 2,"DP" )
DP:
	ICONST
	.dw	212H

; LAST ( -- a )
;	Point to the last name in the name dictionary.
;	HEAD( 4,"LAST" )
LAST:
	ICONST
	.dw	214H

CURR: ; ( -- a )  MM++180712
  ICONST
  .dw 216H

CONT: ; ( -- a )  MM++181003
  ICONST
  .dw 218H

VOCP: ; ( -- a )  MM++180714
  ICONST
  .dw 21AH

CSR: ; ( -- a )  MM++180726
  ICONST
  .dw 21CH

LBB: ; ( -- a )  MM++181208
  ICONST
  .dw 21EH


; DOLIT,0 is used so often, that it should be substituted with ZERO

  HEAD( 1,"0" )
ZERO: ; ( -- 0 )  MM--190208
  ICONST
  .dw 0

; so is DOLIT,1

ONE: ; ( -- 1 )  MM--190208
  ICONST
  .dw 1


; HERE ( -- a )
;	Return the top of the code dictionary.
	HEAD( 4,"HERE" )
HERE:
	INEST
	.dw	DP,AT,EXIT

; PAD ( -- a )
;	Return the address of a temporary buffer.
	HEAD( 3,"PAD" )
PAD:
	INEST
	.dw	HERE,DOLIT,80,PLUS,EXIT

; TIB ( -- a )
;	Return the address of the terminal input buffer.
	HEAD( 3,"TIB" )
TIB:
	ICONST
	.dw	TIBB

;; Numeric output, single precision

; DIGIT ( u -- c )
;	Convert digit u to a character.
;	HEAD( 5,"DIGIT" )
DIGIT:
	cmp	#10,tos
	jl	DIGIT1
	add	#7,tos
DIGIT1:
	add	#"0",tos
	INEXT

; EXTRACT ( n base -- n c )
;	Extract the least significant digit from n.
;	HEAD( 7,"EXTRACT" )
EXTRC:
	INEST
	.dw	ZERO,SWAP,UMMOD
	.dw	SWAP,DIGIT,EXIT

; <# ( -- )
;	Initiate the numeric output process.
	HEAD( 2,"<#" )
BDIGS:
	INEST
	.dw	PAD,HLD,STORE,EXIT

; HOLD ( c -- )
;	Insert a character into the numeric output string.
	HEAD( 4,"HOLD" )
HOLD:
	INEST
	.dw	HLD,AT,ONEM
	.dw	DUPP,HLD,STORE,CSTOR,EXIT

; # ( u -- u )
;	Extract one digit from u and append the digit to output string.
	HEAD( 1,"#" )
DIG:
	INEST
	.dw	BASE,AT,EXTRC,HOLD,EXIT

; #S ( u -- 0 )
;	Convert u until all digits are added to the output string.
	HEAD( 2,"#S" )
DIGS:
	INEST
DIGS1:
	.dw	DIG,DUPP,QBRAN,DIGS2
	.dw	BRAN,DIGS1
DIGS2:
	.dw	EXIT

; SIGN ( n -- )
;	Add a minus sign to the numeric output string.
	HEAD( 4,"SIGN" )
SIGN:
	INEST
	.dw	ZLESS,QBRAN,SIGN1
	.dw	DOLIT,"-",HOLD
SIGN1:
	.dw	EXIT

; #> ( w -- b u )
;	Prepare the output string to be TYPE'd.
	HEAD( 2,"#>" )
EDIGS:
	INEST
	.dw	DROP,HLD,AT
	.dw	PAD,OVER,SUBB,EXIT

; str ( n -- b u )
;	Convert a signed integer to a numeric string.
;	HEAD( 3,"str" )
STR:
	INEST
	.dw	DUPP,TOR,ABSS
	.dw	BDIGS,DIGS,RFROM
	.dw	SIGN,EDIGS,EXIT

; HEX ( -- )
;	Use radix 16 as base for numeric conversions.
	HEAD( 3,"HEX" )
HEX:
	mov	#16,&0x202
	INEXT

; DECIMAL ( -- )
;	Use radix 10 as base for numeric conversions.
	HEAD( 7,"DECIMAL" )
DECIM:
	mov	#10,&0x202
	INEXT

;; Numeric input, single precision

; DIGIT? ( c base -- u t )
;	Convert a character to its numeric value. A flag indicates success.
;	HEAD( 6,"DIGIT?" )
DIGTQ:
	mov	@stack,temp0
	sub	#"0",temp0
	jl	FALSE1
	cmp	#10,temp0
	jl	DIGTQ1
	sub	#7,temp0
; -- bug fix MM-180628 --
  cmp #10,temp0
  jl FALSE1
; -----------------------  
DIGTQ1:
	cmp	tos,temp0
	mov	temp0,0(stack)
	jl	TRUE1
FALSE1:
	clr	tos
	INEXT
TRUE1:
	mov	#-1,tos
	INEXT

; NUMBER? ( a -- n T | a F )
;	Convert a number string to integer. Push a flag on tos.
	HEAD( 7,"NUMBER?" )
NUMBQ:
	INEST
	.dw	BASE,AT,TOR,ZERO,OVER,COUNT
	.dw	OVER,CAT,DOLIT,'$',EQUAL,QBRAN,NUMQ1
	.dw	HEX,SWAP,ONEP,SWAP,ONEM
NUMQ1:
	.dw	OVER,CAT,DOLIT,'-',EQUAL,TOR
	.dw	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
	.dw	QBRAN,NUMQ6
	.dw	ONEM,TOR
NUMQ2:
	.dw	DUPP,TOR,CAT,BASE,AT,DIGTQ
	.dw	QBRAN,NUMQ4
	.dw	SWAP,BASE,AT,STAR,PLUS,RFROM,ONEP
	.dw	DONXT,NUMQ2
	.dw	RAT,SWAP,DROP,QBRAN,NUMQ3
	.dw	NEGAT
NUMQ3:
	.dw	SWAP
	.dw	BRAN,NUMQ5
NUMQ4:
	.dw	RFROM,RFROM,DDROP,DDROP,ZERO
NUMQ5:
	.dw	DUPP
NUMQ6:
	.dw	RFROM,DDROP,RFROM,BASE,STORE,EXIT


;; Terminal output

; BL ( -- 32 )
;	Return 32, the blank character.
	HEAD( 2,"BL" )
BLANK:
	ICONST
	.dw	20H

; SPACE ( -- )
;	Send the blank character to the output device.
	HEAD( 5,"SPACE" )
SPACE:
	INEST
	.dw	BLANK,EMIT,EXIT

; SPACES ( +n -- )
;	Send n spaces to the output device.
	HEAD( 6,"SPACES" )
SPACS:
	INEST
	.dw	ZERO,MAX,TOR,BRAN,CHAR2
CHAR1:
	.dw	SPACE
CHAR2:
	.dw	DONXT,CHAR1,EXIT

; TYPE ( b u -- )
;	Output u characters from b.
	HEAD( 4,"TYPE" )
TYPEE:
	INEST
	.dw	TOR,BRAN,TYPE2
TYPE1:
	.dw	DUPP,CAT,TCHAR,EMIT
	.dw	ONEP
TYPE2:
	.dw	DONXT,TYPE1
	.dw	DROP,EXIT

; CR ( -- )
;	Output a carriage return and a line feed.
	HEAD( 2,"CR" )
CR:
	INEST
	.dw	DOLIT,CRR,EMIT
	.dw	DOLIT,LF,EMIT,EXIT

; do$ ( -- a )
;	Return the address of a compiled string.
;	HEAD( COMPO+3,"do$" )
DOSTR:
	INEST
	.dw	RFROM,RAT,RFROM,COUNT,PLUS
	.dw	ALGND,TOR,SWAP,TOR,EXIT

; $"| ( -- a )
;	Run time routine compiled by $". Return address of a compiled string.
;	HEAD( COMPO+3,"$\"|" )
STRQP:
	INEST
	.dw	DOSTR,EXIT	;force a call to do$

; ."| ( -- )
;	Run time routine of ." . Output a compiled string.
;	HEAD( COMPO+3,".\"|" )
DOTQP:
	INEST
	.dw	DOSTR,COUNT,TYPEE,EXIT

; .R ( n +n -- )
;	Display an integer in a field of n columns, right justified.
	HEAD( 2,".R" )
DOTR:
	INEST
	.dw	TOR,STR,RFROM,OVER,SUBB
	.dw	SPACS,TYPEE,EXIT

; U.R ( u +n -- )
;	Display an unsigned integer in n column, right justified.
	HEAD( 3,"U.R" )
UDOTR:
	INEST
	.dw	TOR,BDIGS,DIGS,EDIGS
	.dw	RFROM,OVER,SUBB,SPACS,TYPEE,EXIT

; U. ( u -- )
;	Display an unsigned integer in free format.
	HEAD( 2,"U." )
UDOT:
	INEST
	.dw	BDIGS,DIGS,EDIGS,SPACE,TYPEE,EXIT

; . ( w -- )
;	Display an integer in free format, preceeded by a space.
	HEAD( 1,"." )
DOT:
	INEST
	.dw	BASE,AT,DOLIT,10,XORR	;?decimal
	.dw	QBRAN,DOT1
	.dw	UDOT,EXIT	;no, display unsigned
DOT1:	.dw	STR,SPACE,TYPEE,EXIT	;yes, display signed

; ? ( a -- )
;	Display the contents in a memory cell.
	HEAD( 1,"?" )
QUEST:
	INEST
	.dw	AT,DOT,EXIT

;; Parsing

; parse ( b u c -- b u delta ; <string> )
;	Scan string delimited by c. Return found string and its offset.
;	HEAD( 5,"parse" )
PARS:
	INEST
	.dw	TEMP,STORE,OVER,TOR,DUPP,QBRAN,PARS8
	.dw	ONEM,TEMP,AT,BLANK,EQUAL,QBRAN,PARS3
	.dw	TOR
PARS1:
	.dw	BLANK,OVER,CAT	;skip leading blanks ONLY
	.dw	SUBB,ZLESS,INVER,QBRAN,PARS2
	.dw	ONEP,DONXT,PARS1
	.dw	RFROM,DROP,ZERO,DUPP,EXIT
PARS2:
	.dw	RFROM
PARS3:
	.dw	OVER,SWAP,TOR
PARS4:
	.dw	TEMP,AT,OVER,CAT,SUBB	;scan for delimiter
	.dw	TEMP,AT,BLANK,EQUAL,QBRAN,PARS5
	.dw	ZLESS
PARS5:
	.dw	QBRAN,PARS6
	.dw	ONEP,DONXT,PARS4
	.dw	DUPP,TOR,BRAN,PARS7
PARS6:
	.dw	RFROM,DROP,DUPP,ONEP,TOR
PARS7:
	.dw	OVER,SUBB,RFROM,RFROM,SUBB,EXIT
PARS8:
	.dw	OVER,RFROM,SUBB,EXIT

; PARSE ( c -- b u ; <string> )
;	Scan input stream and return counted string delimited by c.
;	HEAD( 5,"PARSE" )
PARSE:
	INEST
	.dw	TOR,TIB,INN,AT,PLUS	;current input buffer pointer
	.dw	NTIB,AT,INN,AT,SUBB	;remaining count
	.dw	RFROM,PARS,INN,PSTOR,EXIT

; .( ( -- )
;	Output following string up to next ) .
	HEAD( IMEDD+2,".(" )
DOTPR:
	INEST
	.dw	DOLIT,")",PARSE,TYPEE,EXIT

; ( ( -- )
;	Ignore following string up to next ) . A comment.
	HEAD( IMEDD+1,"(" )
PAREN:
	INEST
	.dw	DOLIT,")",PARSE,DDROP,EXIT

; \ ( -- )
;	Ignore following text till the end of line.
;  HEAD( IMEDD+1,"\" )
;   .dw link
;   .set link = $
;   .db IMEDD+1,92,
  HEAD( IMEDD+1,92 )
BKSLA:
	INEST
	.dw	NTIB,AT,INN,STORE,EXIT

; CHAR ( -- c )
;	Parse next word and return its first character.
	HEAD( 4,"CHAR" )
CHAR:
	INEST
	.dw	BLANK,PARSE,DROP,CAT,EXIT

; TOKEN ( -- a ; <string> )
;	Parse a word from input stream and copy it to name dictionary.
	HEAD( 5,"TOKEN" )
TOKEN:
	INEST
	.dw	BLANK,PARSE,DOLIT,31,MIN
TOKEN1:
.if 0  ;MM-190314 
  ; bug: search fails if HERE returns an unaligned address
	.dw	HERE,DDUP,CSTOR,ONEP
	.dw	SWAP,CMOVE,HERE
	.dw	ZERO,HERE,COUNT,PLUS,CSTOR,EXIT
.else  ; move string to 'HERE ALIGNED" 
  .dw HERE,ALGND,DUPP,TOR,DDUP,CSTOR,ONEP
  .dw	SWAP,CMOVE,RAT
  .dw	ZERO,RFROM,COUNT,PLUS,CSTOR,EXIT
.endif


; WORD ( c -- a ; <string> )
;	Parse a word from input stream and copy it to code dictionary.
	HEAD( 4,"WORD" )
WORDD:
	INEST
	.dw	PARSE,BRAN,TOKEN1

;; Dictionary search

; NAME> ( na -- ca )
;	Return a code address given a name address.
  HEAD( 5,"NAME>" )
NAMET:
	mov.b	@tos+,temp0
	and	#0x1F,temp0
	add	temp0,tos
	inc	tos
	bic	#1,tos
	INEXT

; SAME? ( a a -- a a f \ -0+ )
;	Compare u cells in two strings. Return 0 if identical.
;	HEAD( 5,"SAME?" )
SAMEQ:
	pushs
	mov	2(stack),tos
	mov.b	@tos,tos
SAME1:
	mov	2(stack),temp0
	add	tos,temp0
	mov.b	0(temp0),temp0
	mov	0(stack),temp1
	add	tos,temp1
	mov.b	0(temp1),temp1
	sub	temp1,temp0
	jnz	SAME2
	dec	tos
	jnz	SAME1
	INEXT
SAME2:
	jmp TRUE1


; ID? ( wid na -- f )  MM++180714
; Return true if the word at na is a member of the wordlist wid.
; HEAD( 3,"ID?" )
IDQ:
  INEST
  .dw DUPP,CAT,DOLIT,20H,ANDD
  .dw QBRAN,IDQ1
; tagged word, check if tag = wid
  .dw DOLIT,4,SUBB,AT,CLRB0
  .dw BRAN,IDQ2 
IDQ1:
; untagged word, return true if wid = wid(FORTH) = 0
  .dw DROP,ZERO
IDQ2:
  .dw EQUAL,EXIT


; find-in-wordlist ( a wid -- xt na | a ff )  ;MM++180714
WLFND:
  INEST
  .dw TOR,CNTXT
WLFN1:  ; ( a va )
  .dw OVER,SWAP,FIND,DUPP,ZERO,EQUAL
  .dw QBRAN,WLFN2
  .dw SWAP,RFROM,DDROP,EXIT
WLFN2: ; (a xt na )
  .dw RAT,OVER,IDQ,
  .dw QBRAN,WLFN3
  .dw ROT,RFROM,DDROP,EXIT
WLFN3: ; ( a xt na )
  .dw SWAP,DROP,CELLM
  .dw BRAN,WLFN1


; NAME?	( a -- ca na | a F )
;	Search all context vocabularies for a string.
;	HEAD( 5,"NAME?" )
;NAMEQ:                         ;MM--180714
;	INEST                         ;MM--
;	.dw	CNTXT,AT                  ;MM--  ( a va )
FIND: ;( a va -- ca na | a F )  ;MM++
  INEST                         ;MM++
  .dw AT                        ;MM++  ( a na )
FIND1:
	.dw	DUPP,QBRAN,FIND3	;end of dictionary
	.dw	OVER,AT,OVER,AT,DOLIT,MASKK,ANDD,EQUAL
	.dw	QBRAN,FIND4
	.dw	SAMEQ,QBRAN,FIND2	;match
FIND4:
	.dw	CELLM,AT,BRAN,FIND1
FIND2:
	.dw	SWAP,DROP,DUPP,NAMET,SWAP,EXIT
FIND3:
	.dw	EXIT


; NAME?	( a -- ca na | a F )                                   ;MM~~181124
;	Search all context vocabularies for a string.
;	.dw	WORDD-6
;	HEAD( 5,"NAME?" )
NAMEQ:
	INEST
  .dw VOCP,AT,GETB0,QBRAN,NAMQ1       ; ( a )
   ; search VOC search order
  .dw VOCP,AT,CLRB0,WLFND,QDUP,QBRAN,NAMQ2,EXIT
NAMQ1:
   ; search FORTH search order        ; ( a )
  .dw CONT,AT,QDUP,QBRAN,NAMQ3
  .dw WLFND,QDUP,QBRAN,NAMQ3,EXIT     ; TOSO ?
NAMQ3:
  .dw ZERO
  .dw WLFND,QDUP,QBRAN,NAMQ2,EXIT     ; NOSO ?
NAMQ2:
  .dw DOLIT,NROOT,WLFND,EXIT          ; ROOT ?


; ?CSR ( na|0 -- )
; If the word at na has a ctag <> NSTKY store the wid of the ctag to CSR. Other-
; wise store 0 to CSR (no context switch request).
QCSR:
  INEST
  .dw DUPP,QBRAN,QCSR1
  ; ( na )
  .dw DUPP,CAT,DOLIT,XTAG,ANDD,QBRAN,QCSR2 ; ( na )  
   ; ( na ) XTAG=true ?
  .dw DUPP,DOLIT,4,SUBB,AT,GETB0,QBRAN,QCSR2  ; ( na )
   ; ctag = NSTKY ?
  .dw DOLIT,6,SUBB,AT,DUPP,DOLIT,NSTKY,EQUAL,QBRAN,QCSR1 ;( ctag )
   ; ctag = NSTKY , no context switch ( ctag )
  .dw DROP,EXIT
QCSR2: ;( na )
  .dw DUPP,SUBB ;( 0 )
QCSR1: ; ( ctag|0 )
  .dw CLRB0
  .dw CSR,STORE,EXIT

; ?CS ( -- )
; Do a context switch if CSR <> 0 = wid(ctag)
QCS:
  INEST
  .dw CSR,AT  ; ( wid|0 )
  .dw QDUP,QBRAN,QCS1
  ; set VOCP     ( VOC search order )
  .dw SETB0,BRAN,QCS2 
QCS1:
  ; reset VOCP.0 ( FORTH search order )
  .dw VOCP,AT,CLRB0
QCS2:
  .dw VOCP,STORE
  .dw EXIT


;; Terminal input

; ^H ( bot eot cur -- bot eot cur )
;	Backup the cursor by one character.
;	HEAD( 2,"^H" )
BKSP:
	INEST
	.dw	TOR,OVER,RFROM,SWAP,OVER,XORR
	.dw	QBRAN,BACK1
	.dw	DOLIT,BKSPP,EMIT,ONEM
	.dw	BLANK,EMIT,DOLIT,BKSPP,EMIT
BACK1:
	.dw	EXIT

; TAP ( bot eot cur c -- bot eot cur )
;	Accept and echo the key stroke and bump the cursor.
;	HEAD( 3,"TAP" )
TAP:
	INEST
	.dw	DUPP,EMIT,OVER,CSTOR,ONEP,EXIT

; kTAP ( bot eot cur c -- bot eot cur )
;	Process a key stroke, CR or backspace.
;	HEAD( 4,"kTAP" )
KTAP:
	INEST
	.dw	DUPP,DOLIT,CRR,XORR,QBRAN,KTAP2
	.dw	DOLIT,BKSPP,XORR,QBRAN,KTAP1
	.dw	BLANK,TAP,EXIT
KTAP1:
	.dw	BKSP,EXIT
KTAP2:
	.dw	DROP,SWAP,DROP,DUPP,EXIT

; accept ( b u -- b u )
;	Accept characters to input buffer. Return with actual count.
	HEAD( 6,"ACCEPT" )
ACCEP:
	INEST
	.dw	OVER,PLUS,OVER
ACCP1:
	.dw	DDUP,XORR,QBRAN,ACCP4
	.dw	KEY,DUPP,BLANK,SUBB,DOLIT,95,ULESS
	.dw	QBRAN,ACCP2
	.dw	TAP,BRAN,ACCP1
ACCP2:
	.dw	KTAP
ACCP3:
	.dw	BRAN,ACCP1
ACCP4:
	.dw	DROP,OVER,SUBB,EXIT

; QUERY ( -- )
;	Accept input stream to terminal input buffer.
	HEAD( 5,"QUERY" )
QUERY:
	INEST
	.dw	TIB,DOLIT,80,ACCEP,NTIB,STORE
	.dw	DROP,ZERO,INN,STORE,EXIT

;; Error handling

; QUIT inits return stack. ERROR inits both stacks.

; ERROR ( a -- )
;	Return address of a null string with zero count.
;	HEAD( 5,"ERROR" )
ERROR:
	INEST
	.dw	SPACE,COUNT,TYPEE,DOLIT
	.dw	3FH,EMIT,CR,SPSTO,QUIT


; abort" ( f -- )
;	Run time routine of ABORT" . Abort with a message.
;	HEAD( COMPO+6,"abort\"" )
ABORQ:
	INEST
	.dw	QBRAN,ABOR1	;text flag
	.dw	DOSTR,COUNT,TYPEE,SPSTO,QUIT	;pass error string
ABOR1:
	.dw	DOSTR,DROP,EXIT


;; Text interpreter

; $INTERPRET ( a -- )                                        ;MM~~181124
;	Interpret a word. If failed, try to convert it to an integer.
;	HEAD( 10,"$INTERPRET" )
INTER:
	INEST
;  .dw NAMEQ,QDUP  ;?defined  
  .dw NAMEQ,DUPP,QCSR,QDUP  ;?defined  
	.dw	QBRAN,INTE1
	.dw	AT,DOLIT,COMPO,ANDD	;?compile only lexicon bits
	.dw	ABORQ
	.db 13," compile only"
	.dw	EXECU,QCS,EXIT	;execute defined word  ;
INTE1:
	.dw	NUMBQ	;convert a number
	.dw	QBRAN,INTE2,EXIT
INTE2:
	.dw	ERROR	;error


; [ ( -- )
;	Start the text interpreter.
	HEAD( IMEDD+1,"[" )
LBRAC:
	INEST
	.dw	DOLIT,INTER,TEVAL,STORE,EXIT

; .OK ( -- )
;	Display 'ok' only while interpreting.
;	HEAD( 3,".OK" )
DOTOK:
	INEST
	.dw	DOLIT,INTER,TEVAL,AT,EQUAL
	.dw	QBRAN,DOTO1
	.dw	DOTQP
	.db 3," ok"
DOTO1:	.dw	CR,EXIT


; ?STACK ( -- )
;	Abort if the data stack underflows.
;	HEAD( 6,"?STACK" )
QSTAC:
	INEST
	.dw	DEPTH,ZLESS	;check only for underflow
	.dw	ABORQ
	.db 10," underflow",0
	.dw	EXIT

; EVAL ( -- )
;	Interpret the input stream.
;	HEAD( 4,"EVAL" )
EVAL:
	INEST
EVAL1:
	.dw	TOKEN,DUPP,CAT	;?input stream empty
	.dw	QBRAN,EVAL2
	.dw	TEVAL,ATEXE,QSTAC	;evaluate input, check stack
	.dw	BRAN,EVAL1
EVAL2:
	.dw	DROP,DOTOK,EXIT	;prompt

; QUIT ( -- )
;	Reset return stack pointer and start text interpreter.
	HEAD( 4,"QUIT" )
QUIT:
	INEST
	.dw	RPSTO,LBRAC	;start interpretation
  .dw VOCP,AT,CLRB0,VOCP,STORE  ; reset search context   MM++181124
QUIT1:
	.dw	QUERY,EVAL	;get input
	.dw	BRAN,QUIT1	;continue till error

;; Compiler utilities

; ALLOT ( n -- )
;	Allocate n bytes to the RAM dictionary.
	HEAD( 5,"ALLOT" )
ALLOT:
	INEST
	.dw	DP,PSTOR,EXIT	;adjust code pointer

; IALLOT ( n -- )
;	Allocate n bytes to the code dictionary.
	HEAD( 6,"IALLOT" )
IALLOT:
	INEST
	.dw	CP,PSTOR,EXIT	;adjust code pointer

; I! ( n a -- )
;	Store n to address a in code dictionary.
	HEAD( 2,"I!" )
ISTORE:
	mov	#FWKEY,&FCTL3 ; Clear LOCK
	mov	#FWKEY+WRT,&FCTL1 ; Enable write
;	call	#STORE
	mov.w	@stack+,0(tos)
	pops
	mov	#FWKEY,&FCTL1 ; Done. Clear WRT
	mov	#FWKEY+LOCK,&FCTL3 ; Set LOCK
	INEXT

;--------------------------------------- ;MM++180707
; ?I! ( w a -- )
;	Store w to address a in code dictionary.
; Abort if user dictionary space is full.
QISTOR:
  INEST 
  .dw QFLASH,ISTORE,EXIT
; --------------------------------------

; IERASE ( a -- )
;	Erase a segment at address a.
;	HEAD( 5,"ERASE" )   ;MM--180630
	HEAD( 6,"IERASE" )  ;MM++
IERASE:
	mov	#FWKEY,&FCTL3 ; Clear LOCK
	mov	#FWKEY+ERASE,&FCTL1 ; Enable erase
	clr	0(tos)
	mov	#FWKEY+LOCK,&FCTL3 ; Set LOCK
	pops
	INEXT

; IWRITE ( src dest n -- )
;	Copy n bytes from src to dest.  Dest is in flash memory.
;	HEAD( 5,"WRITE" )    ;MM--
	HEAD( 6,"IWRITE"  )  ;MM++
IWRITE:
	INEST
	.dw	TWOSL,TOR
IWRITE1:
;	.dw	OVER,AT,OVER,ISTORE    ;MM--180707
	.dw	OVER,AT,OVER,QISTOR    ;MM++
	.dw	CELLP,SWAP,CELLP,SWAP
	.dw	DONXT,IWRITE1
	.dw	DDROP,EXIT

; , ( w -- )
;	Compile an integer into the code dictionary.
	HEAD( 1,"," )
COMMA:
	INEST
;	.dw	CP,AT,DUPP,CELLP	;cell boundary                     ;MM--180707
	.dw	CP,AT,QFLASH,DUPP,CELLP	;cell boundary + flash test  ;MM++
	.dw	CP,STORE,ISTORE,EXIT

; call, ( w -- )
;	Compile a call instruction into the code dictionary.
;	HEAD( 5,"call," )
CALLC:
	INEST
	.dw	DOLIT,CALLL,COMMA
	.dw	COMMA,EXIT

; [COMPILE] ( -- ; <string> )
;	Compile the next immediate word into code dictionary.
;	HEAD( IMEDD+9,"[COMPILE]" )
BCOMP:
	INEST
	.dw	TICK,COMMA,EXIT

; COMPILE ( -- )
;	Compile the next address in colon list to code dictionary.
;	HEAD( COMPO+7,"COMPILE" )
COMPI:
	INEST
	.dw	RFROM,DUPP,AT,COMMA	;compile address
	.dw	CELLP,TOR,EXIT	;adjust return address

; LITERAL ( w -- )
;	Compile tos to code dictionary as an integer literal.
	HEAD( IMEDD+7,"LITERAL" )
LITER:
	INEST
	.dw	COMPI,DOLIT,COMMA,EXIT

; $," ( -- )
;	Compile a literal string up to next " .
;	HEAD( 3,"$,\"" )
STRCQ:
	INEST
;   .dw     DOLIT,""""      ;MM-180624 assembles to |DOLIT|0000H|
    .dw     DOLIT,0022H     ;             should be |DOLIT|ASCII(")|
	.dw	WORDD    ;move string to code dictionary
  .dw ZERO  ;MM++ ( a 0 ) no count byte flags
	.dw	STRCQ1,EXIT

; STRCQ1:  ;( a -- ) called from $," or $,,n           ;MM--180713
  STRCQ1:  ;( a f -- )                                 ;MM++
           ;( a 0 -- ) when called from $,"   
           ;( a COMPO|IMEDD|XTAG -- )when called from $,,n
	INEST
;	.dw	DUPP,CAT,TWOSL	;MM-- ;calculate aligned end of string
  .dw OVER,CAT,TWOSL  ;MM++ ( a f cnt )
	.dw	TOR
  .dw OVER,PSTOR      ;MM++ ( store f to count byte at a )
STRCQ2:
	.dw	DUPP,AT,COMMA,CELLP
	.dw	DONXT,STRCQ2
	.dw	DROP,EXIT

;; Structures

; FOR ( -- a )
;	Start a FOR-NEXT loop structure in a colon definition.
	HEAD( IMEDD+3,"FOR" )
FOR:
	INEST
	.dw	COMPI,TOR,BEGIN,EXIT

; BEGIN ( -- a )
;	Start an infinite or indefinite loop structure.
	HEAD( IMEDD+5,"BEGIN" )
BEGIN:
	INEST
	.dw	CP,AT,EXIT

; NEXT ( a -- )
;	Terminate a FOR-NEXT loop structure.
	HEAD( IMEDD+4,"NEXT" )
NEXT:
	INEST
	.dw	COMPI,DONXT,COMMA,EXIT

; UNTIL ( a -- )
;	Terminate a BEGIN-UNTIL indefinite loop structure.
	HEAD( IMEDD+5,"UNTIL" )
UNTIL:
	INEST
	.dw	COMPI,QBRAN,COMMA,EXIT

; AGAIN ( a -- )
;	Terminate a BEGIN-AGAIN infinite loop structure.
	HEAD( IMEDD+5,"AGAIN" )
AGAIN:
	INEST
	.dw	COMPI,BRAN,COMMA,EXIT

; IF ( -- A )
;	Begin a conditional branch structure.
	HEAD( IMEDD+2,"IF" )
IFF:
	INEST
	.dw	COMPI,QBRAN,BEGIN
	.dw	DOLIT,2,IALLOT,EXIT

; AHEAD ( -- A )
;	Compile a forward branch instruction.
;	HEAD( IMEDD+5,"AHEAD" )
AHEAD:
	INEST
	.dw	COMPI,BRAN,BEGIN
	.dw	DOLIT,2,IALLOT,EXIT

; REPEAT ( A a -- )
;	Terminate a BEGIN-WHILE-REPEAT indefinite loop.
	HEAD( IMEDD+6,"REPEAT" )
REPEA:
	INEST
	.dw	AGAIN,BEGIN,SWAP,ISTORE,EXIT

; THEN ( A -- )
;	Terminate a conditional branch structure.
	HEAD( IMEDD+4,"THEN" )
THENN:
	INEST
	.dw	BEGIN,SWAP,QISTOR,EXIT

; AFT ( a -- a A )
;	Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.
	HEAD( IMEDD+3,"AFT" )
AFT:
	INEST
	.dw	DROP,AHEAD,BEGIN,SWAP,EXIT

; ELSE ( A -- A )
;	Start the false clause in an IF-ELSE-THEN structure.
	HEAD( IMEDD+4,"ELSE" )
ELSEE:
	INEST
	.dw	AHEAD,SWAP,THENN,EXIT

; WHILE ( a -- A a )
;	Conditional branch out of a BEGIN-WHILE-REPEAT loop.
	HEAD( IMEDD+5,"WHILE" )
WHILE:
	INEST
	.dw	IFF,SWAP,EXIT

; ABORT" ( -- ; <string> )
;	Conditional abort with an error message.
  .dw link
  .set link = $
  .db IMEDD+6,"ABORT",34,0
ABRTQ:
	INEST
	.dw	COMPI,ABORQ,STRCQ,EXIT

; $" ( -- ; <string> )
;	Compile an inline string literal.
  .dw link
  .set link = $
  .db IMEDD+2,36,34,0
STRQ:
	INEST
	.dw	COMPI,STRQP,STRCQ,EXIT

; ." ( -- ; <string> )
;	Compile an inline string literal to be typed out at run time.
  .dw link
  .set link = $
  .db IMEDD+2,".",34,0
DOTQ:
	INEST
	.dw	COMPI,DOTQP,STRCQ,EXIT

;; Colon compiler

; ?UNIQUE ( a -- a )
;	Display a warning message if the word already exists.
;	HEAD( 7,"?UNIQUE" )
UNIQU:
	INEST
;	.dw	DUPP,NAMEQ           ;?name exists             ;MM--180714
  .dw DUPP,CURR,AT,CLRB0,WLFND  ;?name exists        ;MM++
	.dw	QBRAN,UNIQ1          ;redefinitions are OK
	.dw	DOTQP
	.db 7," reDef "          ;but warn the user
	.dw	OVER,COUNT,TYPEE     ;just in case its not planned
UNIQ1:
	.dw	DROP,EXIT


; $,n ( a -- )
;	Build a new dictionary name using the string at a.
;	HEAD( 3,"$,n" )
SNAME:
	INEST
; .dw ZERO,SWAP ; ( 0 a )  ;MM++180714
  .dw LBB,AT,DOLIT,0E0H,ANDD,SWAP     ;MM-181208
  .dw BRAN,FNAM1              ;MM++
; $,,n ( f a )                ;MM++
; Build a new dictionary name using string a and flags f.
FNAME:
  INEST                       ;MM++
FNAM1:                        ;MM++
  .dw ZERO,LBB,STORE       ;MM++181208   reset LBB
  .dw CP,AT,ALGND,CP,STORE    ; align CP    MM++181004
	.dw	DUPP,CAT	;?null input
	.dw	QBRAN,SNAM1
	.dw	UNIQU	;?redefinition
   ; ( f a ) ? compile a ctag and a wtag  ;MM++180726
  .dw SWAP,CURR,AT,DUPP,GETB0,QBRAN,SNAM2  ; ( a f wtag )
   ; compile ctag
  .dw VOCP,AT,CLRB0,COMMA   ; ( a f wtag )
SNAM2:  ; ( a f wtag )
  .dw DUPP,QBRAN,SNAM3
  ; ( a f wtag ) compile wtag, reset CURR.0
  .dw DUPP,COMMA,CLRB0,CURR,STORE  ; ( a f )
  .dw DOLIT,XTAG,ORR               ; ( a f ) set XTAG bit in f
  .dw BRAN,SNAM0
SNAM3: ; ( a f 0|1 )
  .dw DROP
SNAM0: ; ( a f )
	.dw	LAST,AT,COMMA	    ;save na for vocabulary link
	.dw	CP,AT,LAST,STORE
	.dw	STRCQ1,EXIT       ;fill name field
SNAM1:
	.dw	STRQP
	.db 5," name"	;null input
	.dw	ERROR


; $COMPILE ( a -- )
;	Compile next word to code dictionary as a token or literal.
;	HEAD( 8,"$COMPILE" )
SCOMP:
	INEST
;	.dw	NAMEQ,QDUP  ;?defined    ;MM--181124
  .dw NAMEQ,DUPP,QCSR,QDUP      ;MM++
	.dw	QBRAN,SCOM2
	.dw	AT,DOLIT,IMEDD,ANDD   ;?immediate
	.dw	QBRAN,SCOM1
;	.dw	EXECU,EXIT            ;its immediate, execute          ;MM--181124
	.dw	EXECU,QCS,EXIT        ;its immediate, execute          ;MM++
SCOM1:
;	.dw	COMMA,EXIT            ;its not immediate, compile      ;MM--
  .dw	COMMA,QCS,EXIT        ;its not immediate, compile      ;MM++
SCOM2:
	.dw	NUMBQ                 ;try to convert to number
	.dw	QBRAN,SCOM3
	.dw	LITER,EXIT            ;compile number as integer
SCOM3:	.dw	ERROR           ;error


; OVERT ( -- )
;	Link a new word into the current vocabulary.
;	HEAD( 5,"OVERT" )
OVERT:
	INEST
	.dw	LAST,AT,CNTXT,STORE,EXIT

; ; ( -- )
;	Terminate a colon definition.
	HEAD( IMEDD+COMPO+1,";" )
SEMIS:
	INEST
	.dw	COMPI,EXIT,LBRAC,OVERT,EXIT

; ] ( -- )
;	Start compiling the words in the input stream.
	HEAD( 1,"]" )
RBRAC:
	INEST
	.dw	DOLIT,SCOMP,TEVAL,STORE,EXIT

; : ( -- ; <string> )
;	Start a new colon definition using next word as its name.
	HEAD( 1,":" )
COLON:
	INEST
	.dw	TOKEN,SNAME
COLO1:
  .dw DOLIT,DOLST,CALLC,RBRAC,EXIT

; :NN ( -- xt )                                            ;MM++181004
; Start a headerless colon definition.
  HEAD( 3,":NN" )
CONN:
  INEST
  .dw CP,AT,BRAN,COLO1

;; Defining words

; HEADER ( -- ; <string> )
;	Compile a new array entry without allocating code space.
;	HEAD( 6,"HEADER" )
HEADER:
	INEST
	.dw	TOKEN,SNAME,OVERT
	.dw	DOLIT,DOCON,CALLC,EXIT


; CREATE ( -- ; <string> )
;	Compile a new array entry without allocating code space.
	HEAD( 6,"CREATE" )
CREAT:
	INEST
	.dw	HEADER,DP,AT,COMMA,EXIT

; CONSTANT ( n -- ; <string> )
;	Compile a new constant.
	HEAD( 8,"CONSTANT" )
CONST:
	INEST
	.dw	HEADER,COMMA,EXIT

; VARIABLE ( -- ; <string> )
;	Compile a new variable initialized to 0.
	HEAD( 8,"VARIABLE" )
VARIA:
	INEST
	.dw	CREAT,DOLIT,2,ALLOT,EXIT

;; Tools

; ' ( -- ca )                                                 ;MM~~181207
;	Search context vocabulary for the next word in input stream.
  WTAG( NROOT )
	HEAD( XTAG+1,"'" )
TICK:
	INEST
	.dw	TOKEN,NAMEQ ; defined ?
	.dw	QBRAN,TICK1
	.dw	EXIT        ;yes, push code address
TICK1:
	.dw	ERROR       ;no, error

.if 0
; DUMP( a u -- )
;	Dump u bytes from a, in a formatted manner.
	HEAD( 4,"DUMP" )
DUMP:
	INEST
	.dw	DOLIT,7,TOR	;start count down loop
DUMP1:
	.dw	CR,DUPP,DOLIT,5,UDOTR
	.dw	DOLIT,15,TOR
DUMP2:
	.dw	COUNT,DOLIT,3,UDOTR
	.dw	DONXT,DUMP2	;loop till done
	.dw	SPACE,DUPP,DOLIT,16,SUBB
	.dw	DOLIT,16,TYPEE	;display printable characters
	.dw	DONXT,DUMP1	;loop till done
	.dw	DROP,EXIT
.endif

; .S ( ... -- ... )
;	Display the contents of the data stack.
  WTAG( NROOT )
	HEAD( XTAG+2,".S" )
DOTS:
	INEST
;  .dw DOTS,EXIT
	.dw	CR,DEPTH	;stack depth
	.dw	TOR	;start count down loop
	.dw	BRAN,DOTS2	;skip first pass
DOTS1:
	.dw	RAT,PICK,DOT	;index stack, display contents
DOTS2:
	.dw	DONXT,DOTS1	;loop till done
	.dw	DOTQP
	.db 4," <sp",0
	.dw	EXIT

; >NAME ( ca -- na | F )
;	Convert code address to a name address.
;	HEAD( 5,">NAME" )
TNAME:
	INEST
	.dw	TOR,CNTXT,AT	;vocabulary link
TNAM1:
	.dw	DUPP,QBRAN,TNAM2
	.dw	DUPP,NAMET,RAT,XORR	;compare
	.dw	QBRAN,TNAM2
	.dw	CELLM	;continue with next word
	.dw	AT,BRAN,TNAM1
TNAM2:
	.dw	RFROM,DROP,EXIT


; .ID ( na -- )
;	Display the name at address.
  WTAG( NROOT )
; HEAD( 3,".ID" )
  HEAD( XTAG+3,".ID" )
DOTID:
	INEST
	.dw	COUNT,DOLIT,01FH,ANDD	;mask lexicon bits
	.dw	TYPEE,EXIT


CLRB0: ; ( w1 -- w2 ) Clear bit 0 of w1     ;MM++180725
  bic #1,tos
  INEXT

SETB0: ; ( w1 -- w2 ) set bit 0 of w1       ;MM++180725
  bis #1,tos
  INEXT

GETB0: ; ( w -- 0|1 )   return bit 0 of w     ;MM++190208
  and #1,tos
  INEXT
  


; Return the address of the current context pointer.
QCON: ; ( -- VOCP|CONT )                    ;MM++181207
  INEST
  .dw VOCP,AT,GETB0,QBRAN,QCON1,VOCP,EXIT
QCON1:
  .dw CONT,EXIT

;	Display the name at address na if it's a member of the actual context.
QID: ; ( na -- )                            ;MM++180714~~181207
  INEST
  .dw QCON,AT,CLRB0,OVER,IDQ
  .dw QBRAN,QID1
  .dw SPACE,DOTID,BRAN,QID2
QID1:
  .dw DROP
QID2:
  .dw EXIT


; WORDS	( -- )      ;MM-181118  moved to the ROOT VOC
; Display the names of the top of the actual search order.   MM~~181207
  WTAG( NROOT )
	HEAD( XTAG+5,"WORDS" )
WORDS:
	INEST
  .dw CR,CNTXT,AT	 ; the top of the dictionaries wordlisT
WORS1:
	.dw	QDUP	;?at end of list
	.dw	QBRAN,WORS2
	.dw	DUPP,QID  	      ;display a name
	.dw	CELLM,AT,BRAN,WORS1
WORS2:
	.dw	EXIT


;; Cold boot

; HI ( -- )
;	Display the sign-on message of eForth.
	HEAD( 2,"HI" )
HI:
	INEST
	.dw	CR,DOTQP
;	.db 13,"430eForth43n1"	;model           ;MM--180801  181004
;	.db 16,"430eForth43n1vis",0	;model       ;MM++
;	.db 16,"430eForth43n2vis",0	;model       ;MM~~181118
;	.db 16,"430eForth43n3vis",0	;model       ;MM~~1812..
;	.db 16,"430eForth43n4vis",0	;model       ;MM~~190104
	.db 16,"430eForth43n5vis",0	;model       ;MM~~190131
;	.dw	CR,EXIT   ;MM-180629
	.dw	EXIT

; APP! ( a -- )	Turnkey
;	HEX : APP! 200 ! 1000 IERASE 200 1000 20 IWRITE ;
	HEAD( 4,"APP!" )
APPST:
	INEST
	.dw	TBOOT,STORE,DOLIT,0x1000,IERASE
	.dw	TBOOT,DOLIT,0x1000,DOLIT,0x20
	.dw	IWRITE,EXIT

; -- Flash tools ------------------------                 ;MM-180629

EDM equ 0FFC0H-2  ; top of users dictionary space in the flash memory

; fscan ( -- )  ;MM++180629
; Scan the Flash memory from EDM downwards and set CP to the next free cell
; above the last used one.
; HEAD( 5,"fscan" )
FSCAN:
  INEST
   .dw DOLIT,EDM
FSCN1:
   .dw DOLIT,2,SUBB,DUPP,AT,DOLIT,EM,SUBB
   .dw QBRAN,FSCN1
   .dw DOLIT,2,PLUS,CP,AT,ALGND,MAX,CP,STORE,EXIT  ;

; ?flash ( a -- a )  ;MM++180707
; Abort if user dictionary is full ( a > EDM )
; HEAD( COMPO+6,"?flash" )
QFLASH:
; EDM OVER U< ABORT"  Flash?"
  INEST 
  .dw DOLIT,EDM,OVER,ULESS,ABORQ,
  .db 7," Flash?"
  .dw EXIT

; -- VOCs and ITEMs ------------------------------         ;MM~~190105

; doVP ( -- ) runtime for VOCS 
; HEAD( COMPO+4,"doVP" )
DOVP:
  INEST
; .dw RFROM,AT,CSR,STORE,QCS,EXIT     ;MM~~190208
  .dw RFROM,AT                        ;
DOVP1:                                ;
  .dw CSR,STORE,QCS,EXIT              ;


; ROOT ( -- )                                               ;MM~~191004
; The root VOC for the default and the VOC search order.
  HEAD( IMEDD+4,"ROOT" )
  .set NROOT = link
ROOT:
  INEST
  .dw DOVP,NROOT


; FORTH ( -- )                                              ;MM~~181120
; The FORTH VOC. The default search order is FORTH ROOT.
  HEAD( IMEDD+5,"FORTH" )
  .set NFORTH = link
FORTH:
  INEST
  .dw DOVP,1


; DEFINITIONS ( -- )                                          ;MM~~181207
; Make the actual VOC the current compilation context.
  WTAG(NROOT)
  HEAD( IMEDD+XTAG+11,"DEFINITIONS" )
DEFS:
  INEST
  .dw QCON
DEFS1:
;  .dw AT,CLRB0,CURR,STORE,EXIT                      ;MM~~190129
  .dw AT,CLRB0,CURR,STORE,oDDI,EXIT                 ;


; VOC ( "name" -- )                                  ;MM~~190105
; Create a vocabulary prefix.
  HEAD( 3,"VOC" )
VOC:
  INEST
	.dw	DOLIT,IMEDD,TOKEN,FNAME,DOLIT,DOLST,CALLC,OVERT
	.dw	DOLIT,DOVP,COMMA,LAST,AT,COMMA,EXIT


; ITEM ( -- )                                      ;MM~~190105
; Make the next created word a context switching one that restores the VOC
; context it was created in.
  WTAG( NROOT )
  HEAD( XTAG+4,"ITEM" )
ITEM:
  INEST
  .dw CURR,AT,SETB0,CURR,STORE,EXIT


; .. ( -- ) ; IMMEDIATE                            ;MM~~190105
; Switch back from a VOC context to the default search order.
  WTAG( NROOT )
  HEAD( IMEDD+XTAG+2,".." )
oDDI:
  INEST
; .dw ZERO,CSR,STORE,QCS,EXIT         ;M~~190208
  .dw ZERO,BRAN,DOVP1

.if 0 ; MM-190330
; FIRST ( -- )                                       ;MM~~181207
; Overwrite the top of the Forth search order with the wid of current VOC.
  WTAG( NROOT )
  HEAD( IMEDD+XTAG+5,"FIRST" )
FIRST:
  INEST
  .dw QCON,AT,CLRB0,CONT,STORE,oDDI,EXIT
.else
; ALSO ( -- )                                       ;MM~~181207
; Add the top of the current VOC search order to the Forth search order. Abort 
; with error message if the Forth search order is full.
  WTAG( NROOT )
  HEAD( IMEDD+XTAG+4,"ALSO" )
ALSO:
  INEST
  .dw CONT,AT,ABORQ
  .db 6," ALSO?",0
  .dw QCON,AT,CLRB0,CONT,STORE,oDDI,EXIT
.endif


; STICKY ( – )                                     ;MM++180826  181003
; Make the next created word a context switching ITEM that restores the VOC
; context it was found in.
; NSTKY:
  HEAD( 6,"STICKY" )
  .set NSTKY = link
STKY:
  INEST
  .dw DOLIT,NSTKY,VOCP,STORE,ITEM,EXIT


; ONLY ( -- )
; Reset all context related pointers.   ;MM++190208
  WTAG( NROOT )
  HEAD( IMEDD+XTAG+4,"ONLY" )
ONLY:
  INEST
; .dw CURR,DOLIT,0AH,ZERO,FILL,EXIT     ;MM~~190211
  .dw CONT,DOLIT,08H,ZERO,FILL,EXIT     ;


; COLD	( -- )
;	The hilevel cold start sequence.
	HEAD( 4,"COLD" )
COLD:
	INEST
	.dw	STOIO
	.dw	DOLIT,UZERO,DOLIT,UPP
	.dw	DOLIT,ULAST-UZERO,CMOVE    ;initialize user area
;  .dw CURR,DOLIT,0AH,DOLIT,0,FILL  ; init contexts and CURRENT ;MM~~181207
  .dw ONLY,DEFS
	.dw	TBOOT,ATEXE	;application boot
  .dw FSCAN,CR                                                 ;MM++180629
	.dw	QUIT	;start interpretation

init:                ;mk Put it closer to cold : jmp was out of range.
	mov	#RPP,SP	; set up stack
	mov	#SPP,stack
	clr	tos
	mov.w   #WDTPW+WDTHOLD,&WDTCTL  ; Stop watchdog timer
	bis.b   #041h,&P1DIR	; P1.0/6 output
	jmp	COLD

CTOP:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;
; COLD start moves the following to user variables.

;	.sect	".infoD"  ; CCS: .sect ; naken: .org   ;mk
    .org 01000H ;mk
;   1000-10FF = 256B information memory   ;mk
;   INFO+000h (INFOD):  RAM save area, user variables ;mk
;   INFO+040h (INFOC):  RAM save area ??? ;mk
;   INFO+080h (INFOB):  user interrupt vectors ??? ;mk
;   INFO+0C0h (INFOA):  configuration data - do not use!! ;mk

UZERO:
	.dw	HI	;200H, boot routine
	.dw	BASEE	;202H, BASE
	.dw	0	;204H, tmp
	.dw	0	;206H, >IN
	.dw	0	;208H, #TIB
	.dw	0	;20AH, HLD
	.dw	INTER	;20CH, 'EVAL
; .dw	COLD-6	;20EH, CONTEXT pointer                 ;MM--180713
	.dw	COLD-6	;20EH, dictionary wordlist pointer     ;MM++
	.dw	CTOP+8	;210H, CP; pass ISR
	.dw	DPP	;220H, DP
	.dw	COLD-6	;214H, LAST
  .dw 0       ;216H CURR   ;MM++180712  new CURRENT pointer
  .dw 0       ;218H CONT   ;MM++181003  new CONTEXT pointer
  .dw 0       ;21AH VOCP   ;MM++180713  new VOC pointer
  .dw 0       ;21CH CSR    ;MM++181124  context switch request, wid(voc) or 0
  .dw 0       ;21EH LBB    ;MM++181208  buffer to collect lexicon bits
ULAST:

;===============================================================
;	.sect   ".reset"	; MSP430 RESET Vector  ;mk

; Interrupt vectors are located in the range FFE0-FFFFh.
;       .org 0FFE0h
intvecs: 
;        DC16 VECAREA+00      ; FFE0 - not used
;        DC16  VECAREA+04     ; FFE2 - not used
;        DC16  VECAREA+08     ; FFE4 - IO port P1
;        DC16  VECAREA+12     ; FFE6 - IO port P2
;        DC16  VECAREA+16     ; FFE8 - not used
;        DC16  VECAREA+20     ; FFEA - ADC10
;        DC16  VECAREA+24     ; FFEC - USCI A0/B0 tx, I2C tx/rx
;        DC16  VECAREA+28     ; FFEE - USCI A0/B0 rx, I2C status
;        DC16  VECAREA+32     ; FFF0 - Timer 0_A3
;        DC16  VECAREA+36     ; FFF2 - Timer 0_A3
;        DC16  VECAREA+40     ; FFF4 - Watchdog
;        DC16  VECAREA+44     ; FFF6 - Comparator A
;        DC16  VECAREA+48     ; FFF8 - Timer 1_A3
;        DC16  VECAREA+52     ; FFFA - Timer 1_A3
;        DC16  VECAREA+56     ; FFFC - NMI, osc.fault, flash violation
         .org 0FFFEh
         DC16  init           ; FFFE - Reset

	.end
;===============================================================

